@@ -93,7 +93,7 @@ struct characteristic {
 	char *desc;
 	struct format *format;
 	uint8_t *value;
-	int vlen;
+	size_t vlen;
 };
 
 struct query_data {
@@ -158,6 +158,7 @@ static void gatt_service_free(void *user_data)
 	g_slist_free(gatt->primary);
 	g_attrib_unref(gatt->attrib);
 	g_free(gatt->path);
+	btd_device_unref(gatt->dev);
 	g_free(gatt);
 }
 
@@ -177,14 +178,6 @@ static int characteristic_handle_cmp(gconstpointer a, gconstpointer b)
 	return chr->handle - handle;
 }
 
-static int characteristic_path_cmp(gconstpointer a, gconstpointer b)
-{
-	const struct characteristic *chr = a;
-	const char *path = b;
-
-	return g_strcmp0(chr->path, path);
-}
-
 static int watcher_cmp(gconstpointer a, gconstpointer b)
 {
 	const struct watcher *watcher = a;
@@ -198,18 +191,6 @@ static int watcher_cmp(gconstpointer a, gconstpointer b)
 	return g_strcmp0(watcher->path, match->path);
 }
 
-static inline DBusMessage *invalid_args(DBusMessage *msg)
-{
-	return g_dbus_create_error(msg, ERROR_INTERFACE ".InvalidArguments",
-			"Invalid arguments in method call");
-}
-
-static inline DBusMessage *not_authorized(DBusMessage *msg)
-{
-	return g_dbus_create_error(msg, ERROR_INTERFACE ".NotAuthorized",
-			"Not authorized");
-}
-
 static void append_char_dict(DBusMessageIter *iter, struct characteristic *chr)
 {
 	DBusMessageIter dict;
@@ -255,42 +236,35 @@ static void watcher_exit(DBusConnection *conn, void *user_data)
 }
 
 static int characteristic_set_value(struct characteristic *chr,
-					const uint8_t *value, uint16_t len)
+					const uint8_t *value, size_t vlen)
 {
-	uint8_t *newvalue;
-
-	newvalue = g_try_malloc0(len);
-	if (newvalue == NULL)
+	chr->value = g_try_realloc(chr->value, vlen);
+	if (chr->value == NULL)
 		return -ENOMEM;
 
-	memcpy(newvalue, value, len);
-	g_free(chr->value);
-	chr->vlen = len;
-	chr->value = newvalue;
+	memcpy(chr->value, value, vlen);
+	chr->vlen = vlen;
 
 	return 0;
 }
 
-static void update_watchers(struct primary *prim, struct characteristic *chr)
+static void update_watchers(gpointer data, gpointer user_data)
 {
-	GSList *l;
+	struct watcher *w = data;
+	struct characteristic *chr = user_data;
+	DBusMessage *msg;
 
-	for (l = prim->watchers; l; l = l->next) {
-		DBusMessage *msg;
-		struct watcher *w = l->data;
+	msg = dbus_message_new_method_call(w->name, w->path,
+				"org.bluez.Watcher", "ValueChanged");
+	if (msg == NULL)
+		return;
 
-		msg = dbus_message_new_method_call(w->name, w->path,
-					"org.bluez.Watcher", "ValueChanged");
-		if (msg == NULL)
-			return;
+	dbus_message_append_args(msg, DBUS_TYPE_OBJECT_PATH, &chr->path,
+			DBUS_TYPE_ARRAY, DBUS_TYPE_BYTE,
+			&chr->value, chr->vlen, DBUS_TYPE_INVALID);
 
-		dbus_message_append_args(msg, DBUS_TYPE_OBJECT_PATH, &chr->path,
-				DBUS_TYPE_ARRAY, DBUS_TYPE_BYTE,
-				&chr->value, chr->vlen, DBUS_TYPE_INVALID);
-
-		dbus_message_set_no_reply(msg, TRUE);
-		g_dbus_send_message(connection, msg);
-	}
+	dbus_message_set_no_reply(msg, TRUE);
+	g_dbus_send_message(connection, msg);
 }
 
 static void events_handler(const uint8_t *pdu, uint16_t len,
@@ -301,7 +275,7 @@ static void events_handler(const uint8_t *pdu, uint16_t len,
 	struct primary *prim;
 	GSList *lprim, *lchr;
 	uint8_t opdu[ATT_MAX_MTU];
-	guint handle = att_get_u16((uint16_t *) &pdu[1]);
+	guint handle = att_get_u16(&pdu[1]);
 	uint16_t olen;
 
 	for (lprim = gatt->primary, prim = NULL, chr = NULL; lprim;
@@ -327,10 +301,10 @@ static void events_handler(const uint8_t *pdu, uint16_t len,
 		g_attrib_send(gatt->attrib, opdu[0], opdu, olen,
 						NULL, NULL, NULL);
 	case ATT_OP_HANDLE_NOTIFY:
-		if (characteristic_set_value(chr, pdu + 2, len - 2) < 0)
+		if (characteristic_set_value(chr, &pdu[3], len - 3) < 0)
 			DBG("Can't change Characteristic %0x02x", handle);
 
-		update_watchers(prim, chr);
+		g_slist_foreach(prim->watchers, update_watchers, chr);
 		break;
 	}
 }
@@ -377,8 +351,8 @@ static void connect_cb(GIOChannel *chan, GError *gerr, gpointer user_data)
 		return;
 	}
 
-	atid = gatt_discover_primary(gatt->attrib, 0x0001, 0xffff, primary_cb,
-									gatt);
+	atid = gatt_discover_primary(gatt->attrib, 0x0001, 0xffff, NULL,
+							primary_cb, gatt);
 	if (atid == 0)
 		goto fail;
 
@@ -434,28 +408,14 @@ static DBusMessage *get_characteristics(DBusConnection *conn,
 	return reply;
 }
 
-static DBusMessage *register_watcher(DBusConnection *conn,
-						DBusMessage *msg, void *data)
+static int l2cap_connect(struct gatt_service *gatt, GError **gerr,
+								gboolean listen)
 {
-	const char *sender = dbus_message_get_sender(msg);
-	struct primary *prim = data;
-	struct gatt_service *gatt = prim->gatt;
-	struct watcher *watcher;
-	GError *gerr = NULL;
 	GIOChannel *io;
-	char *path;
-
-	if (!dbus_message_get_args(msg, NULL, DBUS_TYPE_OBJECT_PATH, &path,
-							DBUS_TYPE_INVALID))
-		return invalid_args(msg);
-
-	if (!g_slist_find_custom(prim->chars, path, characteristic_path_cmp))
-		return g_dbus_create_error(msg, ERROR_INTERFACE ".Failed",
-								"Invalid path");
 
 	if (gatt->attrib != NULL) {
 		gatt->attrib = g_attrib_ref(gatt->attrib);
-		goto done;
+		return 0;
 	}
 
 	/*
@@ -463,30 +423,45 @@ static DBusMessage *register_watcher(DBusConnection *conn,
 	 * Configuration it is necessary to poll the server from time
 	 * to time checking for modifications.
 	 */
-	io = bt_io_connect(BT_IO_L2CAP, connect_cb, gatt, NULL, &gerr,
+	io = bt_io_connect(BT_IO_L2CAP, connect_cb, gatt, NULL, gerr,
 			BT_IO_OPT_SOURCE_BDADDR, &gatt->sba,
 			BT_IO_OPT_DEST_BDADDR, &gatt->dba,
 			BT_IO_OPT_PSM, gatt->psm,
 			BT_IO_OPT_SEC_LEVEL, BT_IO_SEC_LOW,
 			BT_IO_OPT_INVALID);
-	if (!io) {
-		DBusMessage *reply;
-		reply = g_dbus_create_error(msg, ERROR_INTERFACE ".Failed",
-							"%s", gerr->message);
-		g_error_free(gerr);
-
-		return reply;
-	}
+	if (!io)
+		return -1;
 
 	gatt->attrib = g_attrib_new(io);
 	g_io_channel_unref(io);
-	gatt->listen = TRUE;
+	gatt->listen = listen;
 
 	g_attrib_set_destroy_function(gatt->attrib, attrib_destroy, gatt);
 	g_attrib_set_disconnect_function(gatt->attrib, attrib_disconnect,
 									gatt);
 
-done:
+	return 0;
+}
+
+static DBusMessage *register_watcher(DBusConnection *conn,
+						DBusMessage *msg, void *data)
+{
+	const char *sender = dbus_message_get_sender(msg);
+	struct primary *prim = data;
+	struct watcher *watcher;
+	GError *gerr = NULL;
+	char *path;
+
+	if (!dbus_message_get_args(msg, NULL, DBUS_TYPE_OBJECT_PATH, &path,
+							DBUS_TYPE_INVALID))
+		return btd_error_invalid_args(msg);
+
+	if (l2cap_connect(prim->gatt, &gerr, TRUE) < 0) {
+		DBusMessage *reply = btd_error_failed(msg, gerr->message);
+		g_error_free(gerr);
+		return reply;
+	}
+
 	watcher = g_new0(struct watcher, 1);
 	watcher->name = g_strdup(sender);
 	watcher->prim = prim;
@@ -510,7 +485,7 @@ static DBusMessage *unregister_watcher(DBusConnection *conn,
 
 	if (!dbus_message_get_args(msg, NULL, DBUS_TYPE_OBJECT_PATH, &path,
 							DBUS_TYPE_INVALID))
-		return invalid_args(msg);
+		return btd_error_invalid_args(msg);
 
 	match = g_new0(struct watcher, 1);
 	match->name = g_strdup(sender);
@@ -518,7 +493,7 @@ static DBusMessage *unregister_watcher(DBusConnection *conn,
 	l = g_slist_find_custom(prim->watchers, match, watcher_cmp);
 	watcher_free(match);
 	if (!l)
-		return not_authorized(msg);
+		return btd_error_not_authorized(msg);
 
 	watcher = l->data;
 	g_dbus_remove_watch(conn, watcher->id);
@@ -540,24 +515,29 @@ static GDBusMethodTable prim_methods[] = {
 static DBusMessage *set_value(DBusConnection *conn, DBusMessage *msg,
 			DBusMessageIter *iter, struct characteristic *chr)
 {
+	struct gatt_service *gatt = chr->prim->gatt;
 	DBusMessageIter sub;
+	GError *gerr = NULL;
 	uint8_t *value;
 	int len;
 
 	if (dbus_message_iter_get_arg_type(iter) != DBUS_TYPE_ARRAY ||
 			dbus_message_iter_get_element_type(iter) != DBUS_TYPE_BYTE)
-		return invalid_args(msg);
+		return btd_error_invalid_args(msg);
 
 	dbus_message_iter_recurse(iter, &sub);
 
 	dbus_message_iter_get_fixed_array(&sub, &value, &len);
 
-	/* FIXME: missing set the new value */
+	if (l2cap_connect(gatt, &gerr, FALSE) < 0) {
+		DBusMessage *reply = btd_error_failed(msg, gerr->message);
+		g_error_free(gerr);
+		return reply;
+	}
 
-	g_free(chr->value);
-	chr->value = g_malloc(len);
-	memcpy(chr->value, value, len);
-	chr->vlen = len;
+	gatt_write_cmd(gatt->attrib, chr->handle, value, len, NULL, NULL);
+
+	characteristic_set_value(chr, value, len);
 
 	return dbus_message_new_method_return(msg);
 }
@@ -589,23 +569,23 @@ static DBusMessage *set_property(DBusConnection *conn,
 	const char *property;
 
 	if (!dbus_message_iter_init(msg, &iter))
-		return invalid_args(msg);
+		return btd_error_invalid_args(msg);
 
 	if (dbus_message_iter_get_arg_type(&iter) != DBUS_TYPE_STRING)
-		return invalid_args(msg);
+		return btd_error_invalid_args(msg);
 
 	dbus_message_iter_get_basic(&iter, &property);
 	dbus_message_iter_next(&iter);
 
 	if (dbus_message_iter_get_arg_type(&iter) != DBUS_TYPE_VARIANT)
-		return invalid_args(msg);
+		return btd_error_invalid_args(msg);
 
 	dbus_message_iter_recurse(&iter, &sub);
 
 	if (g_str_equal("Value", property))
 		return set_value(conn, msg, &sub, chr);
 
-	return invalid_args(msg);
+	return btd_error_invalid_args(msg);
 }
 
 static GDBusMethodTable char_methods[] = {
@@ -835,15 +815,9 @@ static void update_char_value(guint8 status, const guint8 *pdu,
 	struct gatt_service *gatt = current->prim->gatt;
 	struct characteristic *chr = current->chr;
 
-	if (status != 0)
-		goto done;
+	if (status == 0)
+		characteristic_set_value(chr, pdu + 1, len - 1);
 
-	g_free(chr->value);
-
-	chr->vlen = len - 1;
-	chr->value = g_malloc(chr->vlen);
-	memcpy(chr->value, pdu + 1, chr->vlen);
-done:
 	g_attrib_unref(gatt->attrib);
 	g_free(current);
 }
@@ -881,11 +855,10 @@ static void descriptor_cb(guint8 status, const guint8 *pdu, guint16 plen,
 		uint8_t *info = list->data[i];
 		struct query_data *qfmt;
 
-		handle = att_get_u16((uint16_t *) info);
+		handle = att_get_u16(info);
 
 		if (format == 0x01) {
-			sdp_uuid16_create(&uuid, att_get_u16((uint16_t *)
-								&info[2]));
+			sdp_uuid16_create(&uuid, att_get_u16(&info[2]));
 		} else {
 			/* Currently, only "user description" and "presentation
 			 * format" descriptors are used, and both have 16-bit
@@ -972,17 +945,17 @@ static void char_discovered_cb(guint8 status, const guint8 *pdu, guint16 plen,
 		chr = g_new0(struct characteristic, 1);
 		chr->prim = prim;
 		chr->perm = decl[2];
-		chr->handle = att_get_u16((uint16_t *) &decl[3]);
+		chr->handle = att_get_u16(&decl[3]);
 		chr->path = g_strdup_printf("%s/characteristic%04x",
 						prim->path, chr->handle);
 		if (list->len == 7) {
 			sdp_uuid16_create(&chr->type,
-					att_get_u16((uint16_t *) &decl[5]));
+					att_get_u16(&decl[5]));
 		} else
 			sdp_uuid128_create(&chr->type, &decl[5]);
 
 		if (previous_end) {
-			*previous_end = att_get_u16((uint16_t *) decl);
+			*previous_end = att_get_u16(decl);
 		}
 
 		last = chr->handle;
@@ -1279,8 +1252,8 @@ static void primary_cb(guint8 status, const guint8 *pdu, guint16 plen,
 
 		/* Each element contains: attribute handle, end group handle
 		 * and attribute value */
-		start = att_get_u16((uint16_t *) info);
-		end = att_get_u16((uint16_t *) &info[2]);
+		start = att_get_u16(info);
+		end = att_get_u16(&info[2]);
 
 		prim = g_new0(struct primary, 1);
 		prim->gatt = gatt;
@@ -1289,7 +1262,7 @@ static void primary_cb(guint8 status, const guint8 *pdu, guint16 plen,
 
 		if (list->len == 6) {
 			sdp_uuid16_create(&prim->uuid,
-					att_get_u16((uint16_t *) &info[4]));
+					att_get_u16(&info[4]));
 
 		} else if (list->len == 20) {
 			/* FIXME: endianness */
@@ -1320,8 +1293,8 @@ static void primary_cb(guint8 status, const guint8 *pdu, guint16 plen,
 	 * the Error Code is set to Attribute Not Found.
 	 */
 	gatt->attrib = g_attrib_ref(gatt->attrib);
-	gatt->atid = gatt_discover_primary(gatt->attrib,
-				end + 1, 0xffff, primary_cb, gatt);
+	gatt->atid = gatt_discover_primary(gatt->attrib, end + 1, 0xffff, NULL,
+							primary_cb, gatt);
 done:
 	g_attrib_unref(gatt->attrib);
 }
@@ -1331,67 +1304,23 @@ int attrib_client_register(struct btd_device *device, int psm)
 	struct btd_adapter *adapter = device_get_adapter(device);
 	const char *path = device_get_path(device);
 	struct gatt_service *gatt;
-	GError *gerr = NULL;
-	GIOChannel *io;
 	bdaddr_t sba, dba;
 
-	/*
-	 * Registering fake services/characteristics. The following
-	 * paths/interfaces shall be registered after discover primary
-	 * services only.
-	 */
-
 	adapter_get_address(adapter, &sba);
 	device_get_address(device, &dba);
 
 	gatt = g_new0(struct gatt_service, 1);
-	gatt->dev = device;
+	gatt->dev = btd_device_ref(device);
 	gatt->listen = FALSE;
 	gatt->path = g_strdup(path);
 	bacpy(&gatt->sba, &sba);
 	bacpy(&gatt->dba, &dba);
 	gatt->psm = psm;
 
-	gatt_services = g_slist_append(gatt_services, gatt);
-
-	/* FIXME: we should also listen for incoming connections */
-
-	if (load_primary_services(gatt)) {
+	if (load_primary_services(gatt))
 		DBG("Primary services loaded");
-		return 0;
-	}
 
-	if (psm < 0) {
-		/*
-		 * FIXME: when PSM is not given means that L2CAP fixed
-		 * channel shall be used. For this case, ATT CID(0x0004).
-		 */
-
-		DBG("GATT over LE");
-
-		return 0;
-	}
-
-	io = bt_io_connect(BT_IO_L2CAP, connect_cb, gatt, NULL, &gerr,
-					BT_IO_OPT_SOURCE_BDADDR, &sba,
-					BT_IO_OPT_DEST_BDADDR, &dba,
-					BT_IO_OPT_PSM, psm,
-					BT_IO_OPT_SEC_LEVEL, BT_IO_SEC_LOW,
-					BT_IO_OPT_INVALID);
-
-	if (!io) {
-		error("%s", gerr->message);
-		g_error_free(gerr);
-		gatt_service_free(gatt);
-		return -1;
-	}
-
-	gatt->attrib = g_attrib_new(io);
-	g_io_channel_unref(io);
-
-	g_attrib_set_destroy_function(gatt->attrib, attrib_destroy, gatt);
-	g_attrib_set_disconnect_function(gatt->attrib, attrib_disconnect,
-									gatt);
+	gatt_services = g_slist_append(gatt_services, gatt);
 
 	return 0;
 }
